#!/usr/bin/env python
"""Fork LSST repos into a showow GitHub organization."""
from __future__ import print_function

import argparse
import textwrap
import os
from time import sleep
import progressbar
from .. import codetools


def parse_args():
    """Parse command-line arguments"""
    parser = argparse.ArgumentParser(
        prog='github-fork-repos',
        description=textwrap.dedent("""
        Fork LSST into a shadow GitHub organization.
        """),
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog='Part of codekit: https://github.com/lsst-sqre/sqre-codekit')
    parser.add_argument(
        '-u', '--user',
        required=True,
        help='GitHub username')
    parser.add_argument(
        '-o', '--org',
        dest='shadow_org',
        required=True,
        help='Organization to fork repos into')
    parser.add_argument(
        '--token-path',
        default='~/.sq_github_token',
        help='Use a token (made with github-auth) in a non-standard location')
    parser.add_argument(
        '-d', '--debug',
        action='store_true',
        default=os.getenv('DM_SQUARE_DEBUG'),
        help='Debug mode')
    return parser.parse_args()


def main():
    """Fork all repos into shadow org"""
    args = parse_args()

    if args.debug:
        print('You are', args.user)

    ghb = codetools.login_github(token_path=args.token_path)

    # get the organization object
    organization = ghb.organization('lsst')

    # list of all LSST repos
    repos = [g for g in organization.repositories()]
    repo_count = len(repos)

    if args.debug:
        print(repos)

    widgets = ['Forking: ', progressbar.Bar(), ' ', progressbar.AdaptiveETA()]
    pbar = progressbar.ProgressBar(
        widgets=widgets, max_value=repo_count).start()
    repo_idx = 0
    for repo in repos:
        if args.debug:
            print(repo.name)

        repo.create_fork(args.shadow_org)  # NOQA
        sleep(2)
        pbar.update(repo_idx)
        repo_idx += 1


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""Delete all repos in the Github <user>-shadow org."""
from __future__ import print_function
import textwrap
import argparse
import os
from time import sleep
import progressbar
from .. import codetools


def parse_args():
    """Parse command-line arguments"""
    parser = argparse.ArgumentParser(
        prog='github-delete-shadow',
        description=textwrap.dedent("""Delete all repos in the GitHub
            <username>-shadow org"""),
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog='Part of codekit: https://github.com/lsst-sqre/sqre-codekit')
    parser.add_argument(
        '-u', '--user',
        help='GitHub username',
        dest='user',
        required=True)
    parser.add_argument(
        '--token-path',
        default='~/.sq_github_token_delete',
        help='Use a token (made with github-auth) in a non-standard loction')
    parser.add_argument(
        '-d', '--debug',
        action='store_true',
        default=os.getenv('DM_SQUARE_DEBUG'),
        help='Debug mode')
    return parser.parse_args()


def countdown_timer():
    """Show countdown bar"""
    widgets = ['Pause for panic: ', progressbar.ETA(), ' ', progressbar.Bar()]
    pbar = progressbar.ProgressBar(widgets=widgets, max_value=200).start()
    for i in range(200):
        pbar.update(i)
        sleep(0.1)
    pbar.finish()


def main():
    """Delete user shadow org"""
    work = 0
    nowork = 0

    args = parse_args()
    # Deliberately hardcoding the -shadow part due to cowardice
    orgname = '{user}-shadow'.format(user=args.user)

    if args.debug:
        print('org:', orgname)

    ghb = codetools.login_github(token_path=args.token_path)

    # get the organization object
    organization = ghb.organization(orgname)

    # get all the repos
    repos = [g for g in organization.repositories()]

    print('Deleting all repos in', orgname)
    print('Now is the time to panic and Ctrl-C')

    countdown_timer()

    print('Here goes:')

    if args.debug:
        delay = 5
        print(delay, 'second gap between deletions')
        work = 0
        nowork = 0

    for repo in repos:

        if args.debug:
            print('Next deleting:', repo.name, '...',)
            sleep(delay)

        status = repo.delete()

        if status:
            print('ok')
            work += 1
        else:
            print('FAILED - does your token have delete_repo scope?')
            nowork += 1

    print('Done - Succeed:', work, 'Failed:', nowork)
    if work:
        print('Consider deleting your privileged auth token', args.token_path)


if __name__ == '__main__':
    main()



"""Assorted codetools utility functions."""
from __future__ import print_function
# technical debt
# --------------
# - package
# - check explictly for github3 version

import os
import shutil
import tempfile

import gitconfig
from github3 import login
from travisci import TravisCI
from travisci.travisci import BackendError
import urllib3


__all__ = ['login_github', 'eups2git_ref', 'repos_for_team',
           'github_2fa_callback', 'TempDir', 'gitusername', 'gituseremail',
           'get_team_id_by_name', 'get_git_credential_helper',
           'enable_travisci', 'protect']


def login_github(token_path=None, token=None):
    """Log into GitHub using an existing token.

    Parameters
    ----------
    token_path : str, optional
        Path to the token file. The default token is used otherwise.

    token: str, optional
        Literial token string. If specifified, this value is used instead of
        reading from the token_path file.

    Returns
    -------
    gh : :class:`github3.GitHub` instance
        A GitHub login instance.
    """
    if token is None:
        if token_path is None:
            # Try the default token
            token_path = '~/.sq_github_token'
        token_path = os.path.expandvars(os.path.expanduser(token_path))

        if not os.path.isfile(token_path):
            print("You don't have a token in {0} ".format(token_path))
            print("Have you run github-auth?")
            raise EnvironmentError("No token in %s" % token_path)

        with open(token_path, 'r') as fdo:
            token = fdo.readline().strip()

    ghb = login(token=token, two_factor_callback=github_2fa_callback)

    return ghb


def gitusername():
    """
    Returns the user's name from .gitconfig if available
    """
    # pylint: disable=bare-except
    try:
        mygitconfig = gitconfig.GitConfig()
        return mygitconfig['user.name']
    except:
        return None


def gituseremail():
    """
    Returns the user's email from .gitconfig if available
    """

    # pylint: disable=bare-except
    try:
        mygitconfig = gitconfig.GitConfig()
        return mygitconfig['user.email']
    except:
        return None


def github_2fa_callback():
    """
    Prompt for two-factor code
    """
    # http://github3py.readthedocs.org/en/master/examples/two_factor_auth.html
    code = ''
    while not code:
        # The user could accidentally press Enter before being ready,
        # let's protect them from doing that.
        code = input('Enter 2FA code: ')
    return code


def repos_for_team(org, teams=None):
    """Iterate over repos in a GitHub organization that are in the given
    set of teams.

    Parameters
    ----------
    org : class:`github3.github3.orgs.Organization` instance
        The GitHub organization to operate in. Usually created with the
        :meth:`github3.GitHub.organization` method.
    teams : iterable
        A sequence of team names (as strings). If `None` (default) then
        team identity will be ignored and all repos in the organization
        will be iterated over.

    Yields
    ------
    repo : :class:`github3.repos.repo.Repository`
        Yields repositiory instances that pass organization and team criteria.
    """
    if teams is not None:
        teams = set(teams)
    for repo in org.repositories():
        repo_teams = set([t.name for t in repo.teams()])
        if teams is None:
            yield repo
        elif repo_teams.isdisjoint(teams) is False:
            yield repo


def open_repo(org, repo_name):
    """Open a :class:`github3.repos.repo.Repository` instance by name
    in a GitHub organization.

    Parameters
    ----------
    org : class:`github3.github3.orgs.Organization` instance
        The GitHub organization to operate in. Usually created with the
        :meth:`github3.GitHub.organization` method.
    repo_name : str
        Name of the repository (without the organization namespace).
        E.g. `'afw'`.

    Returns
    -------
    repo : :class:`github3.repos.repo.Repository`
        The repository instance.
    """
    for repo in org.repositories():
        if repo.name == repo_name:
            return repo


def get_team_id_by_name(org, team_name, debug=False):
    """Get the ID of a team in a GitHub organization.

    Parameters
    ----------
    org : class:`github3.github3.orgs.Organization` instance
        The GitHub organization to operate in. Usually created with the
        :meth:`github3.GitHub.organization` method.
    team_name : `str`
        Name of the team to find
    debug : `bool`, optional
        Enable debug output.

    Returns
    -------
    team_id : `int` or `None`
        The team ID as an integer, or `None` if `team_name` is the empty
        string.

    Raises
    ------
    `NameError`
        If there is no team with the given name in the supplied organization.
    """

    if team_name == '':
        if debug:
            print("Searching for empty teamname -> None")
        return None  # Special case for empty teams
    teams = org.teams()
    try:
        while True:
            team = teams.next()
            if debug:
                print("Considering team %s with ID %i" % (team.name, team.id))
            if team.name == team_name:
                if debug:
                    print("Match found.")
                return team.id
    except StopIteration:
        raise NameError("No team '%s' in organization '%s'" % (team_name,
                                                               org.login))


def get_git_credential_helper(username, token):
    """Get a string suitable for inclusion in a git config as a credential
    helper, allowing authenticated access without prompting for a password.

    Useful for, e.g., doing a push to Github from a local repo.

    Parameters
    ----------
    username: `str`
        The GitHub username for the authenticated action.
    token: `str`
        The corresponding access token.

    Returns
    -------
    helper : `str`
        A string which is a runnable shell fragment for use in a git config.
    """
    # The initial bang tells git it's about to run a shell fragment.
    #  The rest is a string, which will be evaluated in the shell (so don't
    #  take unsanitized input from the outside world!)
    # It defines a function to swallow any input and put username and password
    #  on separate lines: it substitutes its parameter values into the
    #  username/password values.
    # This means it will work under `git credential fill`, which is the
    #  entire point.
    helper = '!"f() { cat > /dev/null ; echo username=' + username
    helper += ' ; echo password=' + token + ' ; } ; f"'
    return helper


def eups2git_ref(eups_ref,
                 repo,
                 eupsbuild,
                 versiondb='https://raw.githubusercontent.com/lsst/versiondb/master/manifests',  # NOQA pylint: disable=line-too-long
                 debug=None):
    """Provide the eups tag given a git SHA."""
    # Thought of trying to parse the eups tag for the ref, but given
    # that doesn't help with the tag-based versions, might as well
    # look up versiondb for everything

    # eg. https://raw.githubusercontent.com/lsst/versiondb/master/manifests/b1108.txt  # NOQA
    shafile = versiondb + '/' + eupsbuild + '.txt'
    if debug:
        print(shafile)

    # Get the file tying shas to eups versions
    http = urllib3.poolmanager.PoolManager()
    refs = http.request('GET', shafile)
    if refs.status >= 300:
        raise RuntimeError('Failed GET with HTTP code', refs.status)
    reflines = refs.data.splitlines()

    for entry in reflines:
        # Python 2/3 accomodation
        if not isinstance(entry, str):
            entry = str(entry, 'utf-8')
        # skip commented out and blank lines
        if entry.startswith('#'):
            continue
        if entry.startswith('BUILD'):
            continue
        if entry == '':
            continue

        elements = entry.split()
        eupspkg, sha, eupsver = elements[0:3]
        if eupspkg != repo:
            continue
        # sanity check
        if eupsver != eups_ref:
            raise RuntimeError('Something has gone wrong, release file does '
                               'not match manifest', eups_ref, eupsver)
        # get out if we find it
        if debug:
            print(eupspkg, sha, eupsver)

        break

        # We never reach this sanity check, so I am commenting it out.

        # sanity check that our digest looks like a sha1
        # pat = re.compile('\b[0-9a-f]{5,40}\b')
        # mat = pat.match(sha)
        # if not mat:
        #    raise RuntimeError('does not appear to be a sha1 digest', sha)

    return sha


def enable_travisci(repo, token):
    """Enable TravisCI web hook for the Repository, repo.

    :param :class:`Repository <github3.repos.Repository>` repo: Instance
    used to create a TravisCI webhook.
    :param: str token: Literial GitHub token string.
    :returns: bool
    """
    try:
        tci = TravisCI(github_token=token)
        owner_repo = '{0}/{1}'.format(repo.owner.login, repo.name)
        tci.enable_travis_webhook(owner_repo)
    except BackendError:
        return False
    return True


def protect(repo, branch_name='master'):
    """Protect the GitHub repository's branch, branch_name, using
    LSST SQRE's default attributes.

    :param :class:`Repository <github3.repos.Repository>` repo: Instance
    used to protect the branch protection.
    :param str branch_name: The name of the branch.

    :returns: :class:`Protection <github3.repos.branch.Protection>` object.
    """
    branch = repo.branch(branch_name)
    if branch:
        required_status_checks = {
            'include_admins': True,
            'strict': True,
            'contexts': ['continuous-integration/travis-ci']}
        branch.protect(required_status_checks=required_status_checks,
                       enforce_admins=True)
        return branch.protection()


class TempDir(object):
    """ContextManager for temporary directories.

    For example::

        import os
        with TempDir() as temp_dir:
            assert os.path.exists(temp_dir)
        assert os.path.exists(temp_dir) is False
    """

    # pylint: disable=too-few-public-methods
    def __init__(self):
        super(TempDir, self).__init__()
        self._temp_dir = tempfile.mkdtemp()

    def __enter__(self):
        return self._temp_dir

    def __exit__(self, ttype, value, traceback):
        shutil.rmtree(self._temp_dir)
        self._temp_dir = None



#!/usr/bin/env python
"""Utilities for maintaining and asserting LICENSE and COPYRIGHT in
LSST DM repos.
"""

import os
import fnmatch
import re
import time

import requests
import git

# pylint: disable=anomalous-backslash-in-string, invalid-name

comment_pattern = re.compile(
    '(?P<comment_flag>^[#* ]*)(?P<content>[\d\w\s<]*)')


def convert_boilerplate(code_stream):
    """Convert old-style licensing boilerplate (with full GPLv3) to RFC-45
    style.

    Parameters
    ----------
    code_stream : file-like object
        File handle to code.

    Returns
    -------
    converted_text : str
        Converted code text, as a string.
    """
    # Text for the new copyright/license boilerplate
    template_lines = [
        '{comment} See the COPYRIGHT and LICENSE files in the top-level '
        'directory of this\n',
        '{comment} package for notices and licensing terms.\n'
    ]
    lines = code_stream.readlines()
    new_lines = []
    omitting_mode = False
    for line in lines:
        m = comment_pattern.match(line)
        if m is None:
            # Pass through of non-comment characters
            new_lines.append(line)
            continue

        content = m.group('content').lstrip()
        if m is not None and content.startswith('Copyright'):
            # Trigger replacement at start of Copyright block
            # Make it the template's responsibility to add a single space
            # after the comment prefix string
            comment_text = m.group('comment_flag').rstrip()
            for template_line in template_lines:
                new_lines.append(
                    template_line.format(comment=comment_text))
            omitting_mode = True
        elif content.startswith('see <http'):
            # End replacement mode
            omitting_mode = False
        elif omitting_mode is False:
            # Pass through mode
            if line.rstrip() == '#':
                # remove whitespace from comment header/footer for Python
                new_lines.append('#\n')
            else:
                new_lines.append(line)
    return ''.join(new_lines)


def list_development_years(repo):
    """List the years when a repo was actively developed.

    This function does not search all branches; only the history of `repo`'s
    currently active branch.

    Parameters
    ----------
    repo : :class:`git.Repo` instance
        The GitPython repo instance for this repository.

    Returns
    -------
    years : tuple
        The years where a commit occured in this repo's git history.
    """
    years = []
    for commit in repo.iter_commits():
        struct_time = time.gmtime(commit.authored_date)
        years.append(struct_time.tm_year)
    years = list(set(years))
    years.sort()
    return tuple(years)


def write_lsst_license(path, url=None):
    """Write the LSST license into `path`.

    The license is read from a url, which is at
    http://www.lsstcorp.org/LegalNotices/GPLv3License.txt
    by default. However, the source `url` can be modified.

    Parameters
    ----------
    path : str
        Path to write license file into.
    url : str, optional
        URL to find the source license at. Defaults to the lsstcorp.org license
        file by default.
    """
    if url is None:
        url = 'http://www.lsstcorp.org/LegalNotices/GPLv3License.txt'
    licensedoc = requests.get(url)
    with open(path, 'w') as f:
        f.write(licensedoc.text)


def write_default_copyright(path, repo):
    """Write a default COPYRIGHT file according to RFC-45.

    The default copyright is assigned to LSST/AURA. The years are machine-
    supplied from git history.

    We expect authors from other institutions to add their own copyright lines.
    We also expect that the years in this COPYRIGHT file will be
    maintained with a bot.

    Parameters
    ----------
    path : str
        Path to write the copyright file onto.
    repo : :class:`git.Repo` instance
        The GitPython repo instance for this repository.
    """
    template = 'Copyright {y} The LSST DM Developers\n'
    dev_years = list_development_years(repo)
    if len(dev_years) == 1:
        year_str = str(dev_years[0])
    else:
        year_str = '{0}-{1}'.format(min(dev_years), max(dev_years))
    with open(path, 'w') as f:
        f.write(template.format(y=year_str))


def upgrade_repo(github_repo, branch_name, clone_dir, git_helper):
    """Process a repository and upgrade licensing/copyright to RFC-45 style.

    *Note*: the repository is cloned into `clone_dir`, and changes are
    pushed back to the remote on the `branch_name` branch. However, the
    local clone is not cleaned up. Use :class:`codekit.TempDir` to create
    a self-cleaning temporary directory for the clone.

    Parameters
    ----------
    github_repo : obj
        A GitHub repo, generated by `github3`.
    branch_name : str
        Name of the branch to create from master and commit in.
    clone_dir : str
        Directory to clone the repository into.
    git_helper : str
        A magic shell fragment to allow git push to function against the
        repository.  See codetools.get_git_credential_helper().
    """
    # Clone repo
    repo = git.Repo.clone_from(github_repo.clone_url, clone_dir)

    # Create a ticket branch to work in
    ticket_branch = repo.create_head(branch_name)
    ticket_branch.checkout()

    # Covert boilerplate
    patterns = ('*.py', '*.cpp', '*.cc', '*.h')
    for filepath in all_files(clone_dir, patterns=patterns):
        with open(filepath, 'r') as f:
            new_code = convert_boilerplate(f)
        with open(filepath, 'w') as f:
            f.write(new_code)
        repo.index.add([filepath])
    if repo.is_dirty():
        repo.index.commit('Upgrade license and copyright according to RFC-45')

    # Add LSST license file
    license_path = os.path.join(clone_dir, 'LICENSE')
    write_lsst_license(license_path)
    repo.index.add([license_path])
    repo.index.commit('Add LICENSE file according to RFC-45')

    # Add default COPYRIGHT file, sensitive to git timelines
    copyright_path = os.path.join(clone_dir, 'COPYRIGHT')
    write_default_copyright(copyright_path, repo)
    repo.index.add([copyright_path])
    repo.index.commit('Add COPYRIGHT file according to RFC-45')

    # push branch to remote
    assert repo.is_dirty() is False
    remote = repo.remote(name='origin')
    refspec = 'refs/heads/{br}:refs/heads/{br}'.format(br=branch_name)
    # Add our credentials
    cwr = repo.config_writer()
    if not cwr.has_section("credential"):
        cwr.add_section("credential")
    cwr.set("credential", "helper", git_helper)
    cwr.release()
    # https://gitpython.readthedocs.io/en/stable/tutorial.html
    #  suggests that you need to wait/sync or something?
    time.sleep(1)
    try:
        os.sync()
    except AttributeError:
        # Python 2 doesn't expose this.  But set() is safe there.
        pass
    remote.push(refspec=refspec)


def all_files(root, patterns=('*',)):
    # Expand patterns from semicolon-separated string to list
    """Iterate through files matching a pattern.

    Adapted from Receipe 2.16 of the Python Cookbook, 2nd Ed [1].

    Parameters
    ----------
    root : str
        Root directory to process files in.
    patterns : tuple
        An iterable of strings for searching filenames.
        e.g. ``('*.py*, '*.cpp')``.

    Yields
    ------
    path : str
        Path to a file matching the `patterns`.

    References
    ----------
    .. [1] Martelli, A., Ravenscroft, A., & Ascher, D. (2005). *Python
       Cookbook.* O'Reilly Media, Inc..
    """
    for path, _, files in os.walk(root):
        files.sort()
        for name in files:
            for pattern in patterns:
                if fnmatch.fnmatch(name, pattern):
                    yield os.path.join(path, name)
                    break



#!/usr/bin/env python
"""Moves a bunch of Github repos to a team"""
from __future__ import print_function
# Technical Debt
# -------------
# - will need updating to be new permissions model aware

import os
import logging
import argparse
import textwrap
from time import sleep
from .. import codetools


def parse_args():
    """Parse command-line args"""
    parser = argparse.ArgumentParser(
        prog='github-mv-repos-to-team',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=textwrap.dedent("""Move repo(s) from one team to another.

            Note that --from and --to are required "options".

            Examples:

            ./github_mv_repos_to_team.py --from test_ext2 \
                --to test_ext pipe_tasks apr_util
        """),
        epilog='Part of codekit: https://github.com/lsst-sqre/sqre-codekit'
    )

    parser.add_argument(
        'repos', nargs='+',
        help='Names of repos to move')
    parser.add_argument(
        '--from', required=True, dest='oldteam',
        help='Original team name')
    parser.add_argument(
        '--to', required=True, dest='newteam',
        help='Destination team name')
    parser.add_argument(
        '-o', '--org',
        default=None,
        required=True,
        help='Organization to work in')
    parser.add_argument(
        '--token-path',
        default='~/.sq_github_token',
        help='Use a token (made with github-auth) in a non-standard location')
    parser.add_argument(
        '-d', '--debug',
        action='store_true',
        default=os.getenv('DM_SQUARE_DEBUG'),
        help='Debug mode')
    parser.add_argument('--dry-run', action='store_true')

    return parser.parse_args()


def main():
    """Move the repos"""
    # pylint: disable=too-many-branches
    args = parse_args()

    if args.debug:
        print(args)
        logging.getLogger('requests.packages.urllib3')  # NOQA
        stream_handler = logging.StreamHandler()
        logger = logging.getLogger('github3')
        logger.addHandler(stream_handler)
        logger.setLevel(logging.DEBUG)

    ghb = codetools.login_github(token_path=args.token_path)
    if args.debug:
        print(type(ghb))

    org = ghb.organization(args.org)

    newteamid = codetools.get_team_id_by_name(org, args.newteam, args.debug)
    oldteamid = codetools.get_team_id_by_name(org, args.oldteam, args.debug)

    move_me = args.repos
    if args.debug:
        print(len(move_me), 'repos to be moved')

    status = 0
    status2 = 0

    for rnm in move_me:
        repo = args.org + '/' + rnm.rstrip()
        if newteamid:
            # Add team to the repo
            if args.debug or args.dry_run:
                print('Adding', repo, 'to', args.newteam, '...',)

            if not args.dry_run:
                status += org.add_repository(repo, newteamid)
                if status:
                    print('ok')
                else:
                    print('FAILED')

        # remove repo from old team
        # you cannot move out of Owners

        if oldteamid and args.oldteam != 'Owners':
            if args.debug or args.dry_run:
                print('Removing', repo, 'from', args.oldteam, '...',)

            if not args.dry_run:
                status2 += org.remove_repository(repo, oldteamid)

                if status2:
                    print('ok')
                else:
                    print('FAILED')

        # give the API a rest (*snicker*) we don't want to get throttled
        sleep(1)

    if args.debug:
        print(' ')
        print('Added:', status)
        print('Removed:', status2)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""List repositories on Github belonging to organisations, teams, etc.
"""
from __future__ import print_function
# Technical Debt
# --------------

# Known Bugs
# ----------

import argparse
import textwrap
import os
from .. import codetools


def parse_args():
    """Parse command-line arguments"""
    parser = argparse.ArgumentParser(
        prog='github-list-repos',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=textwrap.dedent("""

        List repositories on Github using various criteria

        Examples:

          github_list_repos --org lsst

          github_list_repos --hide 'Data Management' --hide 'Owners' --org lsst

        Note: --mint and --maxt limits are applied after --hide.
        So for example,

          github_list_repos --maxt 0 --hide Owners --org lsst

        returns the list of repos that are owned by no team besides Owners.
        """),
        epilog='Part of codekit: https://github.com/lsst-sqre/sqre-codekit')
    parser.add_argument(
        '-o', '--org',
        dest='organization',
        help='GitHub Organization name',
        required=True)
    parser.add_argument(
        '--hide', action='append',
        help='Hide a specific team from the output')
    parser.add_argument(
        '--mint', type=int, default='0',
        help='Only list repos that have more than MINT teams')
    parser.add_argument(
        '--maxt', type=int,
        help='Only list repos that have fewer than MAXT teams')
    parser.add_argument(
        '--delimiter', default=', ',
        help='Character(s) separating teams in print out')
    parser.add_argument(
        '--token-path',
        default='~/.sq_github_token',
        help='Use a token (made with github-auth) in a non-standard loction')
    parser.add_argument(
        '-d', '--debug',
        action='store_true',
        default=os.getenv('DM_SQUARE_DEBUG'),
        help='Debug mode')
    return parser.parse_args()


def main():
    """List repos and teams"""
    args = parse_args()
    ghb = codetools.login_github(token_path=args.token_path)

    if not args.hide:
        args.hide = []

    org = ghb.organization(args.organization)

    for repo in org.repositories():
        teamnames = [t.name for t in repo.teams()
                     if t.name not in args.hide]
        maxt = args.maxt if (args.maxt is not None and
                             args.maxt >= 0) else len(teamnames)
        if args.debug:
            print("MAXT=", maxt)

        if args.mint <= len(teamnames) <= maxt:
            print(repo.name.ljust(40) + args.delimiter.join(teamnames))


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""Migrate LSST code to use a minimal style of inline boilerplate and
refer to centralzed LICENSE and COPYRIGHT files.

Use lsst-dm to accomplish RFC-45 compliance in the Stack.

Usage
-----

To migrate boilerplate in all repos in the 'Data Management' team in the
lsst organization:

   lsst-bp -u shipitsquirrel --org lsst --team 'Data Management' \
        --branch mybranch

Or you can run lsst-bp on just a single repository:

    lsst-bp -u shipitsqurrel --org lsst --repo afw --branch mybranch

Optionally, the script can be run against repositories forked into a
shadow github organization. Use the github-fork-repos script to do this:

   github-fork-repos -u shipitsquirrel --org shadowy-org
   lsst-bp -u shipitsquirrel --org shadowy-org --ignore-teams \
        --branch mybranch

Note that github-fork-repos does not carry over GitHub team assignments,
so the --team option will not use useful in shadow organizations.
Instead use --ignore-teams to avoid filtering by teams.

Processing Flow
---------------

1. Clones repositories to disk
2. For each repository,

   - modify the boilerplate of all source files to new style
   - make a LICENSE file with GPLv3 text
   - Make a COPYRIGHT file with years determined from git history

3. Push changes up to forks in a new branch.
4. Issue pull requests

Once lsst-bp is run, it shouldn't need to be run again unless
non-compliant code has been added to the stack. lsst-bp is designed to
be run multiple times without adverse effects to compliant code.
"""

from __future__ import print_function
import sys
import argparse
import textwrap
import os
from .. import codetools
from .. import licensing


def parse_args():
    """CL arguments for lsst-bp."""
    parser = argparse.ArgumentParser(
        prog='lsst-bp',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=textwrap.dedent(sys.modules[__name__].__doc__),
        epilog=textwrap.dedent("""
            See DM-4220 and RFC-45 for more information.

            Part of codekit: https://github.com/lsst-sqre/sqre-codekit
            """))
    parser.add_argument(
        '-u', '--user',
        help='GitHub username',
        required=True)
    parser.add_argument(
        '-o', '--org',
        dest='orgname',
        help='GitHub organization, e.g. lsst or a shadow org setup with '
             'lsst-fork-repos',
        required=True)
    parser.add_argument(
        '--repo',
        default=None,
        help='Upgrade the boilerplate of a single repo, rather than all repos '
             'in the organization matching --team')
    parser.add_argument(
        '--branch',
        help='Branch to create and work on',
        required=True)
    parser.add_argument(
        '--team', action='append', default=['Data Management'],
        help='Act on a specific team')
    parser.add_argument(
        '--ignore-teams', action='store_true', default=False,
        help='Ignore filtering by GitHub teams')
    parser.add_argument(
        '--token-path',
        default='~/.sq_github_token',
        help='Token made with github-auth (set if using non-standard path)')
    parser.add_argument(
        '-d', '--debug',
        action='store_true',
        default=os.getenv('DM_SQUARE_DEBUG'),
        help='Debug mode')
    return parser.parse_args()


def main():
    """CLI entrypoint for lsst-bp executable."""
    args = parse_args()

    ghb = codetools.login_github(token_path=args.token_path)
    org = ghb.organization(args.orgname)

    # pylint: disable=redefined-variable-type
    if args.repo is None and args.ignore_teams is True:
        repo_iter = org.repositories()
    elif args.repo is None and args.ignore_teams is False:
        repo_iter = codetools.repos_for_team(org, args.team)
    else:
        # This is actually a list, hence the Pylint disable above.
        repo_iter = [codetools.open_repo(org, args.repo)]

    token = ''
    token_path = os.path.expandvars(os.path.expanduser(args.token_path))
    with open(token_path, 'r') as fdo:
        token = fdo.readline().strip()
    git_helper = codetools.get_git_credential_helper(args.user, token)

    for repo in repo_iter:
        print("Upgrading {0}".format(repo.name))
        with codetools.TempDir() as temp_dir:
            licensing.upgrade_repo(repo, args.branch, temp_dir, git_helper)



#!/usr/bin/env python
"""Generate a github auth token"""
from __future__ import print_function
# technical debt:
# --------------
# - add command line option to override default user
# - add command line option for delete scope

from getpass import getpass
import argparse
import textwrap
import os
import platform
import sys
import github3
from .. import codetools


def parse_args():
    """Parse command line arguments"""
    parser = argparse.ArgumentParser(
        prog='github-auth',
        description=textwrap.dedent("""Generate a GitHub auth token.

           By default this token will not allow you to delete repositories.
           Use the --delete-role flag to create a delete-enabled token.

           By default, regular and delete-enabled tokens will be stored
           in separate locations (~/.sq_github_token vs
           ~/.sq_github_token_delete).
           """),
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog='Part of codekit: https://github.com/lsst-sqre/sqre-codekit')
    parser.add_argument(
        '-u', '--user',
        help='GitHub username',
        dest='user',
        required=True)
    parser.add_argument(
        '--delete-role',
        default=False,
        action='store_true',
        help='Add the delete role to this token')
    parser.add_argument(
        '--token-path',
        default=None,
        help='Save this token to a non-standard path')
    parser.add_argument(
        '-d', '--debug',
        action='store_true',
        default=os.getenv('DM_SQUARE_DEBUG'),
        help='Debug mode')
    return parser.parse_args()


def main():
    """Log in and store credentials"""
    args = parse_args()

    appname = sys.argv[0]
    hostname = platform.node()

    if args.debug:
        print(args.user)
    password = ''

    if args.token_path is None and args.delete_role is True:
        cred_path = os.path.expanduser('~/.sq_github_token_delete')
    elif args.token_path is None and args.delete_role is False:
        cred_path = os.path.expanduser('~/.sq_github_token')
    else:
        cred_path = os.path.expandvars(os.path.expanduser(args.token_path))

    if not os.path.isfile(cred_path):
        print("""
        Type in your password to get an auth token from github
        It will be stored in {0}
        and used in subsequent occasions.
        """.format(cred_path))

        while not password:
            password = getpass('Password for {0}: '.format(args.user))

        note_template = '{app} via github3 on {host} by {user} {creds}'
        note = note_template.format(app=appname,
                                    host=hostname,
                                    user=args.user,
                                    creds=cred_path)
        note_url = 'https://lsst.org/'

        if args.delete_role:
            scopes = ['repo', 'user', 'delete_repo', 'admin:org']
        else:
            scopes = ['repo', 'user']

        auth = github3.authorize(
            args.user, password, scopes, note, note_url,
            two_factor_callback=codetools.github_2fa_callback)

        with open(cred_path, 'w') as fdo:
            fdo.write(auth.token + '\n')
            fdo.write(str(auth.id))

        print('Token written to {0}'.format(cred_path))

    else:
        print("You already have an auth file: {0} ".format(cred_path))
        print("Delete it if you want a new one and run again")
        print("Remember to also remove the corresponding token on Github")


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""Setup Tools Script"""
import os
import codecs
from setuptools import setup, find_packages

PACKAGENAME = 'jmatt-sqre-codekit'
DESCRIPTION = 'LSST Data Management SQuaRE code management tools'
AUTHOR = 'Frossie Economou'
AUTHOR_EMAIL = 'frossie@lsst.org'
URL = 'https://github.com/lsst-sqre/sqre-codekit'
VERSION = '2.1.0'
LICENSE = 'MIT'


def read(filename):
    """Convenience function for includes"""
    full_filename = os.path.join(
        os.path.abspath(os.path.dirname(__file__)),
        filename)
    return codecs.open(full_filename, 'r', 'utf-8').read()


long_description = read('README.md')  # pylint:disable=invalid-name


setup(
    name=PACKAGENAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=long_description,
    url=URL,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    license=LICENSE,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'License :: OSI Approved :: MIT License',
    ],
    keywords='lsst',
    packages=find_packages(exclude=['docs', 'tests*']),
    install_requires=[
        'sqre-github3.py==1.0.0a4',
        'uritemplate.py==2.0.0',
        'requests>=2.8.1,<3.0.0',
        'urllib3==1.12',
        'progressbar2==3.11.0',
        'GitPython==1.0.1',
        'MapGitConfig==1.1',
        'sqre-pytravisci==0.0.3'
    ],
    tests_require=[
        'pytest==3.0.7',
        'pytest-cov==2.4.0',
        'pytest-flake8==0.8.1'],
    # package_data={},
    entry_points={
        'console_scripts': [
            'github-auth = codekit.cli.github_auth:main',
            'github-delete-shadow = codekit.cli.github_delete_shadow:main',
            'github-fork-repos = codekit.cli.github_fork_repos:main',
            'github-list-repos = codekit.cli.github_list_repos:main',
            'github-mv-repos-to-team = codekit.cli.github_mv_repos_to_team:main',  # NOQA
            'github-tag-version = codekit.cli.github_tag_version:main',
            'lsst-bp = codekit.cli.lsst_bp:main',
        ]
    }
)



#!/usr/bin/env python
"""
Use URL to EUPS candidate tag file to git tag repos with official version
"""

# Technical Debt
# --------------
# - sort out the certificate so we don't have to supress warnings
# - completely hide eups-specifics from this file
# - skips non-github repos - can add repos.yaml knowhow to address this
# - worth doing the smart thing for externals? (yes for Sims)
# - deal with authentication version

# Known Bugs
# ----------
# Yeah, the candidate logic is broken, will fix


from __future__ import print_function

import logging
import os
import sys
import argparse
import textwrap
import string
from datetime import datetime
from getpass import getuser
import urllib3
from .. import codetools


def parse_args():
    """Parse command-line arguments"""
    user = getuser()

    parser = argparse.ArgumentParser(
        prog='github-tag-version',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=textwrap.dedent("""

        Tag all repositories in a GitHub org using a team-based scheme

        Examples:
        github-tag-version --org lsst --team 'Data Management' w.2015.33 b1630

        github-tag-version --org lsst --team 'Data Management' \
            --team 'External' --candidate v11_0_rc2 11.0.rc2 b1679

        Note that the access token must have access to these oauth scopes:
            * read:org
            * repo

        The token generated by `github-auth --user` should have sufficient
        permissions.
        """),
        epilog='Part of codekit: https://github.com/lsst-sqre/sqre-codekit'
    )

    # for safety, default to dummy org <user>-shadow
    # will fail for most people but see github_fork_repos in this module
    # on how to get your own

    parser.add_argument('tag')
    parser.add_argument('manifest')
    parser.add_argument(
        '--org',
        default=user + '-shadow')
    parser.add_argument(
        '--team',
        action='append',
        required=True,
        help="team whose repos may be tagged (can specify several times")
    parser.add_argument('--candidate')
    parser.add_argument('--dry-run', action='store_true')
    parser.add_argument(
        '--tagger',
        help='Name of person making the tag - defaults to gitconfig value')
    parser.add_argument(
        '--email',
        help='Email address of tagger - defaults to gitconfig value')
    parser.add_argument(
        '--token-path',
        default='~/.sq_github_token_delete',
        help='Use a token (made with github-auth) in a non-standard location')
    parser.add_argument(
        '--token',
        default=None,
        help='Literal github personal access token string')
    parser.add_argument(
        '-d', '--debug',
        action='store_true',
        default=os.getenv('DM_SQUARE_DEBUG'),
        help='Debug mode')
    parser.add_argument('-v', '--version',
                        action='version', version='%(prog)s 0.5')
    return parser.parse_args()


def main():
    """Create the tag"""
    # pylint: disable=too-many-locals,too-many-nested-blocks,too-many-branches
    # pylint: disable=too-many-statements
    # Although maybe that is a hint that we should break this up...
    args = parse_args()

    orgname = args.org
    version = args.tag

    # if email not specified, try getting it from the gitconfig
    email = args.email
    if email is None:
        email = codetools.gituseremail()
        if email is None:
            sys.exit("Specify --email option")
    if args.debug:
        print("email is " + email)
    # ditto for the name of the tagger
    tagger = args.tagger
    if tagger is None:
        tagger = codetools.gitusername()
        if tagger is None:
            sys.exit("Specify --tagger option")
    if args.debug:
        print("tagger name is " + tagger)

    # The candidate is assumed to be the requested EUPS tag unless
    # otherwise specified with the --candidate option The reason to
    # currently do this is that for weeklies and other internal builds,
    # it's okay to eups publish the weekly and git tag post-facto. However
    # for official releases, we don't want to publish until the git tag
    # goes down, because we want to eups publish the build that has the
    # official versions in the eups ref.

    if args.candidate:
        candidate = args.candidate
    else:
        candidate = args.tag

    eupsbuild = args.manifest  # sadly we need to "just" know this
    message_template = 'Version {v} release from {c}/{b}'
    message = message_template.format(v=version, c=candidate, b=eupsbuild)
    eupspkg_site = 'https://sw.lsstcorp.org/eupspkg/'

    # generate timestamp for github API
    now = datetime.utcnow()
    timestamp = now.isoformat()[0:19] + 'Z'
    if args.debug:
        print(timestamp)

    tagstuff = dict(name=tagger,
                    email=email,
                    date=timestamp)

    if args.debug:
        print(tagstuff)

    ghb = codetools.login_github(token_path=args.token_path, token=args.token)
    if args.debug:
        print(type(ghb))

    if args.debug:
        print("Tagging repos in ", orgname)

    # generate eups-style version
    # eups no likey semantic versioning markup, wants underscores

    # Python 2/3 compatibility
    try:
        # Python 2
        cmap = string.maketrans('.-', '__')
    except AttributeError:
        # Python 3
        cmap = str.maketrans('.-', '__')  # pylint: disable=no-member

    # eups_version = version.translate(map)
    eups_candidate = candidate.translate(cmap)

    # construct url
    eupspkg_taglist = '/'.join((eupspkg_site, 'tags',
                                eups_candidate + '.list'))
    if args.debug:
        print(eupspkg_taglist)

    http = urllib3.poolmanager.PoolManager()

    # supress the certificate warning - technical debt
    urllib3.disable_warnings()  # NOQA
    if args.debug:
        # pylint: disable=fixme
        # FIXME what's going on here? assigning a logger to a package?
        logging.getLogger('requests.packages.urllib3')  # NOQA
        stream_handler = logging.StreamHandler()
        logger = logging.getLogger('github3')
        logger.addHandler(stream_handler)
        logger.setLevel(logging.DEBUG)

    manifest = http.request('GET', eupspkg_taglist)

    if manifest.status >= 300:
        sys.exit("Failed GET")

    entries = manifest.data.splitlines()

    for entry in entries:
        # Python 2/3 accomodation
        if not isinstance(entry, str):
            entry = str(entry, 'utf-8')
        # skip commented out and blank lines
        if entry.startswith('#'):
            continue
        if entry.startswith('EUPS'):
            continue
        if entry == '':
            continue

        # extract the repo and eups tag from the entry
        (upstream, _, eups_tag) = entry.split()
        if args.debug:
            print(upstream, eups_tag)

        # okay so we still have the data dirs on gitolite
        # for now, just skip them and record them.
        # question is should they be on different tagging scheme anyway?
        # at this point these are: afwdata, astrometry_net_data qserv_testdata

        repo = ghb.repository(orgname, upstream)

        # if the repo is not in github skip it for now
        # see TD
        if not hasattr(repo, 'name'):
            print('!!! SKIPPING', upstream, (60 - len(upstream)) * '-')
            continue

        if not sum(1 for _ in repo.teams()):
            print('!!! repo has NO teams -- SKIPPING', upstream,
                  (45 - len(upstream)) * '-')
            continue

        for team in repo.teams():
            if team.name in args.team:
                if args.debug or args.dry_run:
                    print(repo.name.ljust(40), 'found in', team.name)
                sha = codetools.eups2git_ref(eups_ref=eups_tag,
                                             repo=repo.name,
                                             eupsbuild=eupsbuild,
                                             debug=args.debug)
                if args.debug or args.dry_run:
                    print('Will tag sha: {sha} as {v} (was {t})'.format(
                        sha=sha, v=version, t=eups_tag))

                if not args.dry_run:
                    try:
                        # create_tag() returns a Tag object on success or None
                        # on failure
                        tag = repo.create_tag(tag=version,
                                              message=message,
                                              sha=sha,
                                              obj_type='commit',
                                              tagger=tagstuff,
                                              lightweight=False)
                        if tag is None:
                            raise RuntimeError('failed to create git tag')

                    except Exception as exc:  # pylint: disable=broad-except
                        print('OOPS: -------------------')
                        print(str(exc))
                        print('OOPS: -------------------')

            else:
                if args.debug:
                    print('No action for', repo.name,
                          'belonging to', team.name)


if __name__ == '__main__':
    main()
